/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2008-2009 Bull S.A.S.
 * Contact: cmi@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * --------------------------------------------------------------------------
 * $Id: ClusterViewManagerFactory.java 2343 2009-06-02 13:11:26Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.controller.factory;

import java.lang.reflect.Constructor;
import java.net.MalformedURLException;
import java.net.UnknownHostException;
import java.util.List;

import net.jcip.annotations.ThreadSafe;

import org.ow2.cmi.admin.MBeanUtils;
import org.ow2.cmi.config.CMIConfig;
import org.ow2.cmi.config.CMIConfigException;
import org.ow2.cmi.config.CMIConfigurator;
import org.ow2.cmi.config.JNDIConfig;
import org.ow2.cmi.controller.common.ClusterViewManager;
import org.ow2.cmi.controller.common.IConfig;
import org.ow2.cmi.reference.ServerId;
import org.ow2.cmi.reference.ServerRef;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Factory handling the configuration of a cluster view manager and creating one.
 * @author Loris Bouzonnet
 */
@ThreadSafe
public final class ClusterViewManagerFactory {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(ClusterViewManagerFactory.class);

    /**
     * Implementation of the client cluster view manager.
     */
    private static String clientClusterViewManagerClassname =
        "org.ow2.cmi.controller.client.ClientClusterViewManagerImpl";

    /**
     * Singleton.
     */
    private static ClusterViewManagerFactory factory;

    /**
     * The configuration of the created cluster view manager.
     */
    private CMIConfig<?> cmiConfig;

    /**
     * The cluster view manager that is created by this factory.
     */
    private ClusterViewManager clusterViewManager;

    /**
     * Singleton: use the static method.
     */
    private ClusterViewManagerFactory() {
        // Do nothing
    }

    /**
     * @return a factory to create and configure an instance of cluster view manager
     */
    public static synchronized ClusterViewManagerFactory getFactory() {
        if(factory == null) {
            factory = new ClusterViewManagerFactory();
        }
        return factory;
    }

    /**
     * @return the cluster view manager that is created by this factory.
     */
    public ClusterViewManager getClusterViewManager() {
        return clusterViewManager;
    }

    /**
     * Returns the instance (singleton) of cluster view manager.
     * This one is created and configured at the first call.
     * @return the instance (singleton) of cluster view manager.
     * @throws ClusterViewManagerFactoryException if a cluster view manager cannot be constructed
     */
    public synchronized ClusterViewManager create() throws ClusterViewManagerFactoryException {
        // Checks if an instance already exists
        if(clusterViewManager == null) {
            // Check that configuration is ready
            if(cmiConfig == null || cmiConfig.getClusterViewManagerFactoryConfig() == null) {
                logger.error("No configuration");
                throw new ClusterViewManagerFactoryException("No configuration");
            }
            ClusterViewManagerFactoryConfig<?> factoryConfig = cmiConfig.getClusterViewManagerFactoryConfig();
            if(factoryConfig instanceof org.ow2.cmi.controller.factory.server.ClusterViewManagerFactoryConfig) {
                String classname = ((org.ow2.cmi.controller.factory.server.ClusterViewManagerFactoryConfig) factoryConfig)
                .getServerClusterViewManagerClassname();

                // Loads the implementation of ServerClusterViewManager via CMIConfig
                Class<?> serverClusterViewManagerClass;
                try {
                    serverClusterViewManagerClass =
                        CMIConfig.class.getClassLoader().loadClass(classname);
                } catch (ClassNotFoundException e) {
                    logger.error("Cannot load the class implementing a server-side manager.");
                    throw new ClusterViewManagerFactoryException(
                            "Cannot load the class implementing a server-side manager.", e);
                }
                logger.debug("The ServerClusterViewManager is: {0}", serverClusterViewManagerClass.getName());

                // Create an instance of this implementation
                try {
                    clusterViewManager = (ClusterViewManager) serverClusterViewManagerClass.newInstance();
                } catch (Exception e) {
                    logger.error("Cannot create the instance (singleton) of ServerClusterViewManager", e);
                    throw new ClusterViewManagerFactoryException(
                            "Cannot get an instance of ServerClusterViewManager", e);
                }
            } else {
                logger.debug("Constructing a ClientClusterviewManager...");
                String protocol =
                    ((org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig) factoryConfig).getProtocol();
                String initialContextFactoryName =
                    ((org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig) factoryConfig)
                        .getInitialContextFactoryName();
                List<String> clusterViewProviderUrls =
                    ((org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig) factoryConfig)
                        .getClusterViewProviderUrls();
                List<ServerRef> serverRefs;
                try {
                    serverRefs = CMIConfigurator.extractServerRefs(protocol, clusterViewProviderUrls);
                } catch (CMIConfigException e) {
                    logger.error("Unable to extract some server refs from the provider URL " + clusterViewProviderUrls, e);
                    throw new ClusterViewManagerFactoryException(
                            "Unable to extract some server refs from the provider URL " + clusterViewProviderUrls, e);
                }
                String localRegistry =
                    ((org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig) factoryConfig).getLocalRegistryUrl();
                ServerId registryId = null;
                if(localRegistry != null) {
                    try {
                        registryId =
                            new ServerId(MBeanUtils.getMBeanDomainName(),
                                    MBeanUtils.getMBeanServerName(), protocol, localRegistry);
                    } catch (MalformedURLException e) {
                        logger.error("The following provider URL is malformed {0}", localRegistry, e);
                        throw new ClusterViewManagerFactoryException(
                                "The following provider URL is malformed " + localRegistry, e);
                    } catch (UnknownHostException e) {
                        logger.error("The host name of the provider URL {0} cannot be resolved", localRegistry, e);
                        throw new ClusterViewManagerFactoryException(
                                "The host name of the provider URL " + localRegistry + " cannot be resolved", e);
                    }
                }
                Class<?> clientClusterViewManagerClass;
                try {
                    clientClusterViewManagerClass = getClass().getClassLoader().loadClass(clientClusterViewManagerClassname);
                    Constructor<?> constructor =
                        clientClusterViewManagerClass.getConstructor(String.class, String.class, List.class, ServerId.class);
                    clusterViewManager =
                        (ClusterViewManager) constructor.newInstance(initialContextFactoryName, protocol, serverRefs, registryId);
                } catch (Exception e) {
                    logger.error("Unable to create a client cluster view manager for the class "
                            + clientClusterViewManagerClassname, e);
                    throw new ClusterViewManagerFactoryException(
                            "Unable to create a client cluster view manager for the class "
                            + clientClusterViewManagerClassname, e);
                }
            }
        }
        return clusterViewManager;
    }

    /**
     * @return the configuration of the cluster view manager.
     */
    public IConfig getConfig() {
        return cmiConfig.getClusterViewManagerFactoryConfig().getConfig();
    }

    /**
     * @return the JNDI configuration.
     */
    public JNDIConfig getJNDIConfig() {
        return cmiConfig.getJNDIConfig();
    }

    /**
     * Destroy the linked cluster view manager.
     */
    public synchronized void destroy() {
        if(clusterViewManager != null) {
            clusterViewManager.destroy();
            this.clusterViewManager = null;
        }
    }

    /**
     * @return true if this factory has not a linked cluster view manager
     */
    public boolean isDestroyed() {
        return clusterViewManager == null;
    }

    /**
     * Set the configuration to use when creating a cluster view manager.
     * @param cmiConfig the configuration
     */
    public synchronized void configure(final CMIConfig<?> cmiConfig) {
        if(this.cmiConfig == null) {
            this.cmiConfig = cmiConfig;
        }
    }

}
